import contextlib
import logging
from collections import deque
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import folders
from mstrio.helpers import IServerError
from mstrio.object_management import PredefinedFolders
from mstrio.types import ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.helper import (
    fetch_objects_async,
    get_default_args_from_func,
    get_temp_connection,
)
from mstrio.utils.resolvers import (
    get_folder_id_from_params_set,
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import objects as objects_processors

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


# This root Project folder is hardcoded in every Environment
ROOT_PROJECT_FOLDER_ID = "D43364C684E34A5F9B2F9AD7108F7828"


def list_folders(
    connection: "Connection",
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    include_subfolders: bool = False,
    parent_folder: 'Folder | tuple[str] | list[str] | str | None' = None,
    **filters,
) -> list["Folder"] | list[dict]:
    """Get a list of folders - either all folders in a specific project or all
    folders that are outside of projects, called configuration-level folders.
    The list of configuration-level folders includes folders such as users, user
    groups, databases, etc. which are not project-specific. If you pass
    any project-related parameter, you get folders in that project;
    if not, then you get configuration-level folders.

    Note:
        Id of project is not taken directly from `Connection` object, so you
        have to specify it explicitly.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        include_subfolders (bool): if True, all subfolders are included in the
            result. Default is False.
            Note that using this option will result in a large number of
                objects being fetched, as it will fetch all folders available.
        parent_folder (Folder | tuple | list | str, optional): Folder object or
            ID or name or path specifying the folder.
            The path has to be provided in the following format:
                if it's inside of a project, start with a Project Name:
                    /MicroStrategy Tutorial/Public Objects/Metrics
                if it's a root folder, start with `CASTOR_SERVER_CONFIGURATION`:
                    /CASTOR_SERVER_CONFIGURATION/Users
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version', 'acg',
             'owner', 'hidden',
          'ext_type']

    Returns:
        list of `Folder` objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
        assert_id_exists=False,
        no_fallback_from_connection=True,
    )

    validate_owner_key_in_filters(filters)

    # Project is validated only if project was specified in arguments -
    # otherwise fetch is performed from a non-project area.
    temp_conn = get_temp_connection(connection, proj_id)

    if not parent_folder:
        objects = fetch_objects_async(
            temp_conn,
            folders.list_folders,
            folders.list_folders_async,
            limit=limit,
            chunk_size=1000,
            project_id=proj_id,
            filters=filters,
        )
    else:
        parent_folder_id = get_folder_id_from_params_set(
            temp_conn, proj_id, folder=parent_folder
        )

        with config.temp_verbose_disable():
            parent_folder = Folder(connection=temp_conn, id=parent_folder_id)
            objects = parent_folder.get_contents(
                include_subfolders=include_subfolders,
                limit=limit,
                to_dictionary=True,
                type=8,
            )

    if include_subfolders:
        recursive_objects = []

        with config.temp_verbose_disable():
            new_limit = limit - len(objects) if limit else None
            for obj in objects:
                folder = Folder(connection=temp_conn, id=obj.get('id'))
                recursive_objects += folder.get_contents(
                    to_dictionary=True, include_subfolders=True, limit=new_limit, type=8
                )

        objects += recursive_objects

    objects = objects[:limit]

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(temp_conn, objects)


def get_my_personal_objects_contents(
    connection: "Connection",
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
) -> list:
    """Get contents of `My Personal Objects` folder in a specific project.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.

    Returns:
        list of objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    objects = folders.get_my_personal_objects_contents(connection, proj_id).json()
    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


def get_predefined_folder_contents(
    connection: "Connection",
    folder_type: PredefinedFolders,
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    **filters,
) -> list:
    """Get contents of a pre-defined Strategy One folder in a specific project.
    Available values for `folder_type` are stored in enum `PredefinedFolders`.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`
        folder_type (enum): pre-defined folder type. Available values are
            stored in enum `PredefinedFolders`.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        to_dictionary (bool, optional): If True returns dicts, by default
            (False) returns objects.
        limit (int): limit the number of elements returned. If `None` (default),
            all objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'date_created', 'date_modified', 'acg']

    Returns:
        list of objects or list of dictionaries
    """

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    objects = fetch_objects_async(
        connection,
        folders.get_predefined_folder_contents,
        folders.get_predefined_folder_contents_async,
        limit=limit,
        chunk_size=1000,
        folder_type=folder_type.value,
        project_id=proj_id,
        filters=filters,
    )

    if to_dictionary:
        return objects
    else:
        from mstrio.utils.object_mapping import map_objects_list

        return map_objects_list(connection, objects)


def get_folder_id_from_path(
    connection: "Connection",
    path: str | tuple[str] | list[str],
) -> str:
    """Get folder id from folder path.

    Args:
        connection (Connection): Strategy One connection object returned by
            `connection.Connection()`
        path (str): path of the Folder
            the path has to be provided in the following format:
                if it's inside of a project, example:
                    /MicroStrategy Tutorial/Public Objects/Metrics
                if it's a root folder, example:
                    /CASTOR_SERVER_CONFIGURATION/Users

    Returns:
        Folder id.
    """

    def get_err_msg(f_name: str) -> str:
        return (
            f"Couldn't find folder with given name {f_name} "
            f"while exploring the path. Check if provided path is correct."
        )

    if isinstance(path, str):
        if path:
            if path[0] == '/':
                path = path[1:]
            if path[-1] == '/':
                path = path[:-1]
        folders_in_path = path.split('/') if path else []
    else:
        folders_in_path = path

    if not folders_in_path:
        raise ValueError(get_err_msg("<EMPTY>"))

    try:
        project_id = (
            # FYI: EXPECT project ID if project-level...
            get_project_id_from_params_set(connection, project_name=folders_in_path[0])
            if folders_in_path[0] != 'CASTOR_SERVER_CONFIGURATION'
            # ... `None` only otherwise
            else None
        )
    except ValueError as err:
        if "Could not uniquely identify the Project by it's name" in str(err):
            raise ValueError(get_err_msg(folders_in_path[0])) from err
        raise err

    with connection.temporary_project_change(project_id):
        if project_id:
            with config.temp_verbose_disable():
                folder = Folder(connection=connection, id=ROOT_PROJECT_FOLDER_ID)
        else:
            castor_id = (
                folders.get_predefined_folder_id(connection=connection, folder_type=39)
                .json()
                .get('id')
            )
            with config.temp_verbose_disable():
                folder = Folder(connection=connection, id=castor_id)

        for f_name in folders_in_path[1:]:
            temp_folder = next(
                (f for f in folder.get_subfolders() if f.name == f_name),
                None,
            )

            if not temp_folder:
                raise ValueError(get_err_msg(f_name))

            with config.temp_verbose_disable():
                folder = Folder(connection=connection, id=temp_folder.id)

        return folder.id


def _folder_bfs_traversal(start_queue: deque["Folder"]):
    queue = start_queue.copy()
    while queue:
        current_folder = queue.popleft()
        yield current_folder

        with contextlib.suppress(IServerError):
            subfolders = current_folder.get_subfolders()
            queue.extend(subfolders)


def _folder_dfs_traversal(start_stack: list["Folder"]):
    stack = start_stack.copy()
    while stack:
        current_folder = stack.pop()
        yield current_folder

        with contextlib.suppress(IServerError):
            subfolders = current_folder.get_subfolders()
            stack.extend(reversed(subfolders))


class Folder(Entity, CopyMixin, MoveMixin, DeleteMixin):
    """Object representation of Strategy One Folder object.

    Attributes:
        connection: MSTR Connection object
        id: ID of the folder
        name: name of the folder
        description: description of the folder
        type: the type of the folder (based on EnumDSSXMLObjectTypes)
        subtype: the subtype of the folder (based on EnumDSSXMLObjectSubTypes)
        date_created: the date/time at which the folder was first saved into the
            metadata
        date_modified: the date/time at which the folder was last saved into the
            metadata
        version: the version number this folder is currently carrying
        acl: an array of access control entry objects
        owner: User object, the owner of the object
        acg: access rights, EnumDSSXMLAccessRightFlags
        ext_type: object extended type, EnumDSSExtendedType
        ancestors: list of ancestor folders
        abbreviation: folder's abbreviation
        hidden: specifies whether the folder is hidden
        icon_path: folder icon path
        view_media: view media settings
        certified_info: CertifiedInfo object, certification status, time of
            certification, and information about the certifier (currently only
            for document and report)
        contents: contents of folder
        path: folder path string built from `ancestors`
    """

    _FROM_DICT_MAP = {**Entity._FROM_DICT_MAP, 'owner': User.from_dict}

    _API_PATCH: dict = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put')
    }

    _OBJECT_TYPE = ObjectTypes.FOLDER
    _SIZE_LIMIT = 10000000  # this sets desired chunk size in bytes

    def __init__(
        self,
        connection: "Connection",
        id: str | None = None,
        name: str | None = None,
        path: tuple[str] | list[str] | str | None = None,
    ):
        """Initialize folder object by its identifier.

        Note:
            Providing only `name` as identifier can be significantly
            slower, based on amount of folders in the environment, and
            does not guarantee finding correct folder when there are many with
            the same name. This will initialize first found name match.
            Use ID or Path to guarantee uniqueness.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`.
            id (str, optional): Identifier of a pre-existing folder containing
                the required data.
            path (list | tuple | str, optional): Path of the pre-existing
                folder. The path has to be provided in the following format:
                    if it's inside of a project, start with a Project Name:
                        /MicroStrategy Tutorial/Public Objects/Metrics
                    if it's a root folder, start with
                    `CASTOR_SERVER_CONFIGURATION`:
                        /CASTOR_SERVER_CONFIGURATION/Users
            name (str, optional): name of folder.
        """
        if not id:
            if name or path:
                id = get_folder_id_from_params_set(
                    connection,
                    connection.project_id,
                    folder_name=name,
                    folder_path=path,
                )
            else:
                raise ValueError("Either 'id', 'name' or 'path' has to be provided.")

        super().__init__(connection, id, name=name)

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        parent: 'Folder | tuple[str] | list[str] | str | None' = None,
        parent_id: str | None = None,
        parent_name: str | None = None,
        parent_path: tuple[str] | list[str] | str | None = None,
        description: str | None = None,
    ) -> "Folder":
        """Create a new folder in a folder selected within connection object
        by providing its name, id of parent folder and optionally description.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            name (str): name of a new folder.
            parent (Folder | tuple | list | str, optional): Folder object or ID
                or name or path specifying the folder. May be used instead of
                `parent_id`, `parent_name` or `parent_path`.
            parent_id (str, optional): ID of a folder.
            parent_name (str, optional): Name of a folder.
            parent_path (str, optional): Path of the folder.
                The path has to be provided in the following format:
                    if it's inside of a project, start with a Project Name:
                        /MicroStrategy Tutorial/Public Objects/Metrics
                    if it's a root folder, start with
                    `CASTOR_SERVER_CONFIGURATION`:
                        /CASTOR_SERVER_CONFIGURATION/Users
            description (str, optional): optional description of a new folder.

        Returns:
            newly created folder
        """
        connection._validate_project_selected()

        par_id = get_folder_id_from_params_set(
            connection,
            connection.project_id,
            folder=parent,
            folder_id=parent_id,
            folder_name=parent_name,
            folder_path=parent_path,
        )

        response = folders.create_folder(connection, name, par_id, description).json()
        if config.verbose:
            logger.info(
                f"Successfully created folder named: '{response.get('name')}' "
                f"with ID: '{response.get('id')}'"
            )
        return cls.from_dict(source=response, connection=connection)

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ) -> None:
        """Alter the folder properties.

        Args:
            name: folder name
            description: folder description
            hidden: Specifies whether the folder is hidden
            comments: long description of the folder
            owner: (str, User, optional): owner of the folder
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        default_dict = get_default_args_from_func(func)
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def get_contents(
        self,
        to_dictionary: bool = False,
        include_subfolders: bool = False,
        limit: int | None = None,
        **filters,
    ) -> list:
        """Get contents of a folder. It can contain other folders or different
        kinds of objects.

        Args:
            to_dictionary (bool, optional): If True returns dicts, by default
                (False) returns objects.
            include_subfolders (bool, optional): If True returns contents of all
                subfolders as well. False by default.
                Note that using this option may result in a large number of
                objects being fetched, especially if coupled with filters.
            limit (int, optional): Limit the number of elements returned. If
                `None` (default), all objects are returned.
            **filters: Available filter parameters: ['name', 'id', 'type',
                'subtype', 'date_created', 'date_modified', 'version',
                'acg', 'owner', 'ext_type']

        Returns:
            Contents as Python objects (when `to_dictionary` is `False` (default
            value)) or contents as dictionaries otherwise.
        """
        queue = deque([self.id])
        objects = []

        validate_owner_key_in_filters(filters)

        while queue and (limit is None or len(objects) < limit):
            current_id = queue.popleft()
            current_objects = fetch_objects_async(
                self.connection,
                folders.get_folder_contents,
                folders.get_folder_contents_async,
                limit=limit,
                chunk_size=1000,
                id=current_id,
                filters=filters,
            )
            objects.extend(current_objects)

            if include_subfolders and (limit is None or len(objects) < limit):
                current_objects = fetch_objects_async(
                    self.connection,
                    folders.get_folder_contents,
                    folders.get_folder_contents_async,
                    limit=limit,
                    chunk_size=1000,
                    id=current_id,
                    filters={'type': 8},
                )
                child_folders = [
                    child for child in current_objects if child.get('type') == 8
                ]
                queue.extend(child.get('id') for child in child_folders)

        objects = objects[:limit]
        if to_dictionary:
            return objects
        else:
            from mstrio.utils.object_mapping import map_objects_list

            return map_objects_list(self.connection, objects)

    def get_subfolders(self, to_dictionary: bool = False) -> 'list[Folder | dict]':
        """Get direct subfolders of the folder.

        Args:
            to_dictionary (bool, optional): If True returns list of dicts, by
                default (False) returns list of Folder class instances.

        Returns:
            List of direct subfolders as Folder objects or dictionaries.
        """
        return self.get_contents(
            type=ObjectTypes.FOLDER,
            include_subfolders=False,
            to_dictionary=to_dictionary,
        )

    def traversal(
        self,
        yield_self: bool = True,
        dfs: bool = False,
    ):
        """Generate traversal specifically for this folder's structure,
        starting with a folder itself.

        Note:
            This is a Generator method yielding folders one by one in an
            efficient way, making REST requests along the way but allowing for
            interactions with already gathered folders in the meantime.

            Please see Example section below for instructions on how to use it.

        Args:
            yield_self (bool): if `True` (default), the traversal starts with
                the folder itself. If `False`, only its children and their
                descendants are yielded.
            dfs (bool): if `True`, depth-first search is used traversing as deep
                in folder's children children as possible before going to next
                child. If `False` (default), breadth-first search is used, going
                through all direct children of a folder before going level
                deeper.

        Example:
            >>> from mstrio.connection import Connection
            >>> from mstrio.object_management import Folder
            >>>
            >>> conn = Connection(...)
            >>>
            >>> root_folder = Folder(connection=conn, id="<some-id>")
            >>> for folder in root_folder.traversal():
            ...     print(folder.name)
        """

        with contextlib.suppress(IServerError):
            if dfs:
                yield from _folder_dfs_traversal(
                    [self] if yield_self else list(reversed(self.get_subfolders()))
                )
            else:
                yield from _folder_bfs_traversal(
                    deque([self] if yield_self else self.get_subfolders())
                )

    def __str__(self):
        start = super().__str__()
        return f"{start} and path: '{self.path}'"

    @property
    def path(self) -> str:
        """Get full path of the folder as string."""
        return self.location

    @staticmethod
    def traverse_folders(folders_list: 'list[Folder]', dfs: bool = False):
        """Generate traversal for an explicitly provided list of folders.

        Note:
            This is a Generator method yielding folders one by one in an
            efficient way, making REST requests along the way but allowing for
            interactions with already gathered folders in the meantime.

            Please see Example section below for instructions on how to use it.

        Args:
            folders_list (list of Folder): list of folders to traverse
            dfs (bool): if `True`, depth-first search is used traversing as deep
                in folder's children children as possible before going to next
                child. If `False` (default), breadth-first search is used, going
                through all direct children of a folder before going level
                deeper.

        Example:
            >>> from mstrio.connection import Connection
            >>> from mstrio.object_management import Folder, list_folders
            >>>
            >>> conn = Connection(...)
            >>>
            >>> folders = list_folders(conn, include_subfolders=False)
            >>> for folder in Folder.traverse_folders(folders):
            ...     print(folder.name)
        """

        if dfs:
            yield from _folder_dfs_traversal(folders_list)
        else:
            yield from _folder_bfs_traversal(deque(folders_list))



import logging
from typing import TYPE_CHECKING

import pandas as pd
import requests
from tqdm.auto import tqdm

from mstrio import config
from mstrio.api import reports as reports_api
from mstrio.api.schedules import get_contents_schedule
from mstrio.connection import Connection
from mstrio.distribution_services.schedule import Schedule
from mstrio.modeling import Prompt
from mstrio.object_management.search_operations import SearchPattern, full_search
from mstrio.users_and_groups.user import User
from mstrio.utils.cache import CacheSource, ContentCacheMixin
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import (
    CertifyMixin,
    CopyMixin,
    DeleteMixin,
    Entity,
    MoveMixin,
    ObjectSubTypes,
    ObjectTypes,
    PromptMixin,
    VldbMixin,
)
from mstrio.utils.filter import Filter
from mstrio.utils.helper import (
    camel_to_snake,
    exception_handler,
    fallback_on_timeout,
    find_object_with_name,
    get_parallel_number,
    get_total_count_of_objects,
    key_fn_for_sort_object_properties,
    response_handler,
)
from mstrio.utils.parser import Parser
from mstrio.utils.related_subscription_mixin import RelatedSubscriptionMixin
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.response_processors import reports as reports_processors_api
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.version_helper import meets_minimal_version
from mstrio.utils.vldb_mixin import ModelVldbMixin

if TYPE_CHECKING:
    from mstrio.object_management.folder import Folder
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


def list_reports(
    connection: Connection,
    name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    folder: 'Folder | tuple[str] | list[str] | str | None' = None,
    folder_id: str | None = None,
    folder_name: str | None = None,
    folder_path: tuple[str] | list[str] | str | None = None,
    **filters,
) -> list[type['Report']] | list[dict]:
    """Get list of Report objects or dicts with them.
    Optionally filter reports by specifying 'name'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name = ?onny will return Sonny and Tonny

    Args:
        connection: Strategy One connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of reports being searched
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Report objects
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Contains. Possible values are available
            in ENUM mstrio.object_management.SearchPattern.
            Default value is BEGIN WITH (4).
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        folder (Folder | tuple | list | str, optional): Folder object or ID or
            name or path specifying the folder. May be used instead of
            `folder_id`, `folder_name` or `folder_path`.
        folder_id (str, optional): ID of a folder.
        folder_name (str, optional): Name of a folder.
        folder_path (str, optional): Path of the folder.
            The path has to be provided in the following format:
                if it's inside of a project, start with a Project Name:
                    /MicroStrategy Tutorial/Public Objects/Metrics
                if it's a root folder, start with `CASTOR_SERVER_CONFIGURATION`:
                    /CASTOR_SERVER_CONFIGURATION/Users
        **filters: Available filter parameters: ['id', 'type', 'subtype',
            'date_created', 'date_modified', 'version', 'owner', 'ext_type',
            'view_media', 'certified_info']

    Returns:
        list with Report objects or list of dictionaries
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )
    validate_owner_key_in_filters(filters)

    objects_ = full_search(
        connection,
        object_types=ObjectTypes.REPORT_DEFINITION,
        project=proj_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        root=folder,
        root_id=folder_id,
        root_name=folder_name,
        root_path=folder_path,
        **filters,
    )
    reports = [
        item for item in objects_ if Report._is_subtype_supported(item['subtype'])
    ]
    if to_dictionary:
        return reports
    return [
        Report.from_dict(
            source=report_dict, connection=connection, with_missing_value=True
        )
        for report_dict in reports
    ]


class Report(
    Entity,
    CertifyMixin,
    CopyMixin,
    MoveMixin,
    DeleteMixin,
    ContentCacheMixin,
    VldbMixin,
    PromptMixin,
    RelatedSubscriptionMixin,
):
    """Access, filter, publish, and extract data from in-memory reports.

    Create a Report object to load basic information on a report dataset.
    Specify subset of report to be fetched through `Report.apply_filters()` and
    `Report.clear_filters()`. Fetch dataset through `Report.to_dataframe()`
    method.

    _CACHE_TYPE is a variable used by ContentCache class for cache filtering
    purposes.

    Attributes:
        connection: Strategy One connection object returned by
            `connection.Connection()`.
        id: Identifier of a pre-existing report containing the required
            data.
        name: Report name
        description: Report description
        abbreviation: Report abbreviation
        instance_id: Identifier of an instance if report instance has been
            already initialized, NULL by default.
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: owner User object
        view_media: View media information
        ancestors: List of ancestor folders
        certified_info: Information whether report is certified or not,
            CertifiedInfo object
        sql: SQL View of the Report
        attributes: List of attributes
        metrics: List of metrics
        page_by_attributes: List of attributes selected for Page By
        attr_elements: All attributes elements of report
        page_by_elements: Elements of attributes selected for Page By in the
            report. The IDs are in the terse format used for page selection,
            which is different than the one used in the attr_elements field.
        current_page_by: Attribute elements selected for Page By in the report
        selected_attributes: IDs of filtered attributes
        selected_metrics: IDs of filtered metrics
        selected_attr_elements: IDs of filtered attribute elements
        dataframe: content of a report extracted into a Pandas `DataFrame`
        prompts: List of report prompts
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.REPORT_DEFINITION
    _OBJECT_SUBTYPES = [
        ObjectSubTypes.REPORT_GRID,
        ObjectSubTypes.REPORT_GRAPH,
        ObjectSubTypes.REPORT_ENGINE,
        ObjectSubTypes.REPORT_DATAMART,
        ObjectSubTypes.REPORT_GRID_AND_GRAPH,
        ObjectSubTypes.REPORT_TRANSACTION,
        ObjectSubTypes.REPORT_HYPER_CARD,
    ]
    _CACHE_TYPE = CacheSource.Type.REPORT
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'owner': User.from_dict,
        'certified_info': CertifiedInfo.from_dict,
    }
    _SIZE_LIMIT = 10_000_000  # this sets desired chunk size in bytes

    _API_PATCH: dict = {
        (
            'name',
            'description',
            'abbreviation',
            'hidden',
            'folder_id',
            'owner',
        ): (
            objects_processors.update,
            'partial_put',
        )
    }
    _API_GET_PROMPTS = staticmethod(reports_api.get_report_prompts)
    _API_PROMPT_GET_INSTANCE = staticmethod(reports_api.report_instance)
    _API_PROMPT_GET_OBJ_STATUS = staticmethod(reports_processors_api.get_report_status)
    _API_PROMPT_GET_PROMPTED_INSTANCE = staticmethod(reports_api.get_prompted_instance)
    _API_PROMPT_ANSWER_PROMPTS = staticmethod(reports_api.answer_report_prompts)

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
        instance_id: str | None = None,
        parallel: bool = True,
        progress_bar: bool = True,
    ):
        """Initialize an instance of a report.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            id (str): Identifier of a pre-existing report containing
                the required data.
            name (str): Name of the report to be looked up if ID is not provided
            instance_id (str): Identifier of an instance if report instance has
                been already initialized, NULL by default.
            parallel (bool, optional): If True (default), utilize optimal number
                of threads to increase the download speed. If False, this
                feature will be disabled.
            progress_bar(bool, optional): If True (default), show the download
                progress bar.
        """

        connection._validate_project_selected()

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            report = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_reports,
                search_pattern=SearchPattern.EXACTLY,
            )
            super().__init__(
                connection,
                report['id'],
                instance_id=instance_id,
                parallel=parallel,
                progress_bar=progress_bar,
            )
        elif id == '':
            raise ValueError("Report ID cannot be an empty string.")
        else:
            super().__init__(
                connection,
                id,
                instance_id=instance_id,
                parallel=parallel,
                progress_bar=progress_bar,
            )

        # Report exposes VLDB property methods through both VldbMixin and
        # ModelVldbMixin. Due to name confusion, the Modeling Service methods
        # are prefixed with 'model_'.
        self._model_vldb = ModelVldbMixin()
        self._model_vldb._MODEL_VLDB_API = {
            'GET_ADVANCED': reports_api.get_vldb_settings,
            'GET_APPLICABLE': reports_api.get_applicable_vldb_settings,
        }
        self._model_vldb.connection = self.connection
        self._model_vldb.id = self.id

        self.model_list_vldb_settings = self._model_vldb.list_vldb_settings

    def _init_variables(self, default_value, **kwargs):
        super()._init_variables(default_value=default_value, **kwargs)
        self._instance_id = kwargs.get("instance_id")
        self._parallel = kwargs.get("parallel", True)
        self._initial_limit = 1000
        self._progress_bar = kwargs.get("progress_bar", True) and config.progress_bar
        self._cross_tab = False
        self._cross_tab_filter = {}
        self._subtotals = {}
        self._dataframe = None
        self._attr_elements = None
        self._page_by_elements = {}
        self._sql = None
        self._current_page_by = []
        self._valid_page_by_elements = []

        self._attributes = []
        self._metrics = []
        self._page_by_attributes = []
        self.__definition_retrieved = False
        self.__filter = None

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ):
        """Alter Report properties.

        Args:
            name: new name of the Report
            description: new description of the Report
            abbreviation: new abbreviation of the Report
            hidden: Specifies whether the Report is hidden
            comments (str, optional): long description of the Report
            owner: (str, User, optional): owner of the Report
        """
        if isinstance(owner, User):
            owner = owner.id
        func = self.alter
        args = func.__code__.co_varnames[: func.__code__.co_argcount]
        defaults = func.__defaults__  # type: ignore
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {}
        for property_key in default_dict.keys():
            if local[property_key] is not None:
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    def to_dataframe(
        self,
        limit: int | None = None,
        page_element_id: str | list[str] | dict[str, str] | None = None,
        prompt_answers: list[Prompt] | None = None,
    ) -> pd.DataFrame:
        """Extract contents of a report instance into a Pandas `DataFrame`.

        Note:
            If the report has page-by attributes and `page_element_id` is not
            provided, the first valid combination of page-by elements will be
            used.

        Args:
            limit (None or int, optional): Used to control data extract
                behavior. By default (None) the limit is calculated
                automatically, based on an optimized physical size of one chunk.
                Setting limit manually will force the number of rows per chunk.
                Depending on system resources, a higher limit (e.g. 50,000) may
                reduce the total time required to extract the entire dataset.
            page_element_id (str, list[str] or dict[str, str], optional): ID of
                the attribute elements chosen for Page By, for example,
                `'h4;8D679D3511D3E4981000E787EC6DE8A4'`.
                If passed as a list, the attribute elements should be listed in
                the same order as the attributes in the `page_by_attributes`
                field. If passed as a dictionary, the keys should be the
                attribute IDs and the values should be the attribute
                element IDs.
            prompt_answers (None or list of Prompts, optional): List of Prompt
                class objects answering the prompts of the report. Only needed
                if the report has prompts.

        Returns:
            Pandas Data Frame containing the report contents.
        """
        if limit:
            self._initial_limit = limit

        attr_ids = [a['id'] for a in self.page_by_attributes]
        if isinstance(page_element_id, str):
            page_element_id = [page_element_id]
        if isinstance(page_element_id, dict):
            page_element_id = [page_element_id.get(attr_id, "") for attr_id in attr_ids]

        if page_element_id:
            page_element_id = [
                str(el_id).partition(';')[0] + ';' + attr
                for el_id, attr in zip(page_element_id, attr_ids)
            ]

        if self._instance_id is None or page_element_id != self._current_page_by:
            self._current_page_by = page_element_id
            res = self.__initialize_report(self._initial_limit)
        else:
            # try to get first chunk from already initialized instance of report
            # if not possible, initialize new instance
            try:
                res = self.__get_chunk(
                    instance_id=self._instance_id, offset=0, limit=self._initial_limit
                )
            except requests.HTTPError:
                res = self.__initialize_report(self._initial_limit)

        _instance = res.json()
        self._instance_id = _instance['instanceId']

        # Answer prompts if provided
        if prompt_answers:
            json = {"prompts": [prompt.to_dict() for prompt in prompt_answers]}
            reports_api.answer_report_prompts(
                connection=self._connection,
                report_id=self.id,
                instance_id=self._instance_id,
                body=json,
                project_id=self.project_id,
            )
            # Get the instance results again, as they weren't generated
            # properly until the prompts were answered
            _instance = reports_api.report_instance_id(
                connection=self.connection,
                report_id=self.id,
                instance_id=self._instance_id,
                offset=0,
                limit=self._initial_limit,
            ).json()

        if not self._current_page_by and self.valid_page_by_elements:
            self._current_page_by = self.get_selected_page_by_elements(
                self.valid_page_by_elements[0]
            )

        # Check status. At this point the instance should be ready w/ data.
        # If there are outstanding prompts, the status will be 2,
        # i.e. com.microstrategy.webapi.EnumDSSXMLStatus.DssXmlStatusPromptXML
        # https://www2.microstrategy.com/producthelp/Current/ReferenceFiles/reference/constant-values.html#com.microstrategy.webapi.EnumDSSXMLStatus.DssXmlStatusMsgID
        if _instance['status'] == 2:
            raise ValueError(
                f"No data available for report \"{self.name}\". "
                "There are unanswered prompts."
            )
        # Gets the pagination totals from the response object
        paging = _instance['data']['paging']

        # initialize parser and process first response
        p = Parser(response=_instance, parse_cube=False)
        p.parse(response=_instance)

        # If there are more rows to fetch, fetch them
        if paging['current'] != paging['total']:
            if not limit:
                limit = max(
                    1000,
                    int((self._initial_limit * self._SIZE_LIMIT) / len(res.content)),
                )
            # Count the number of additional iterations
            it_total = int((paging['total'] - self._initial_limit) / limit) + (
                (paging['total'] - self._initial_limit) % limit != 0
            )

            if self._parallel and it_total > 1:
                threads = get_parallel_number(it_total)
                with FuturesSessionWithRenewal(
                    connection=self._connection, max_workers=threads
                ) as session:
                    fetch_pbar = tqdm(
                        desc="Downloading",
                        total=it_total + 1,
                        disable=not self._progress_bar or not config.verbose,
                    )
                    future = self.__fetch_chunks_future(
                        session, paging, self._instance_id, limit
                    )
                    fetch_pbar.update()
                    for i, f in enumerate(future, start=1):
                        response = f.result()
                        if not response.ok:
                            response_handler(response, "Error getting report contents.")
                        fetch_pbar.update()
                        fetch_pbar.set_postfix(
                            rows=str(
                                min(self._initial_limit + i * limit, paging['total'])
                            )
                        )
                        p.parse(response.json())
                    fetch_pbar.close()
            else:
                self.__fetch_chunks(p, paging, it_total, self._instance_id, limit)

        # return parsed data as a data frame
        self._dataframe = p.dataframe

        # filter dataframe if report had crosstabs and filters were applied
        if self._cross_tab_filter != {}:
            if self._cross_tab_filter['metrics'] is not None:
                # drop metrics columns from dataframe
                metr_names = [
                    el['name']
                    for el in list(
                        filter(
                            lambda x: x['id'] not in self._cross_tab_filter['metrics'],
                            self.metrics,
                        )
                    )
                ]
                self._dataframe = self._dataframe.drop(metr_names, axis=1)

            if self._cross_tab_filter['attr_elements'] is not None:
                # create dict of attributes and elements to iterate through
                attr_dict = {}
                for attribute in self._cross_tab_filter['attr_elements']:
                    key = attribute[:32]
                    attr_dict.setdefault(key, []).append(attribute[33:])
                # initialize indexes series for filter
                indexes = pd.Series([False] * len(self._dataframe))

                # logical OR for filtered attribute elements
                for attribute in attr_dict:
                    attr_name = list(
                        filter(
                            lambda x, attr=attribute: x['id'] in attr, self.attributes
                        )
                    )[0]['name']
                    elements = attr_dict[attribute]
                    indexes = indexes | self._dataframe[attr_name].isin(elements)
                # select dataframe indexes with
                self._dataframe = self._dataframe[indexes]

            if self._cross_tab_filter['attributes'] is not None:
                attr_names = [
                    el['name']
                    for el in list(
                        filter(
                            lambda x: x['id']
                            not in self._cross_tab_filter['attributes'],
                            self.attributes,
                        )
                    )
                ]
                # filtering out attribute forms columns
                to_be_removed = []
                to_be_added = []
                for attr in attr_names:
                    forms = [
                        column
                        for column in self._dataframe.columns
                        if column.startswith(attr + '@')
                    ]
                    if forms:
                        to_be_removed.append(attr)
                        to_be_added.extend(forms)
                for elem in to_be_removed:
                    attr_names.remove(elem)
                attr_names.extend(to_be_added)
                # drop filtered out columns
                self._dataframe = self._dataframe.drop(attr_names, axis=1)
        return self._dataframe

    def __fetch_chunks_future(self, future_session, pagination, instance_id, limit):
        """Fetch added rows from this object instance from the Intelligence
        Server."""
        return [
            reports_api.report_instance_id_coroutine(
                future_session,
                report_id=self._id,
                instance_id=instance_id,
                offset=_offset,
                limit=limit,
            )
            for _offset in range(self._initial_limit, pagination['total'], limit)
        ]

    def __fetch_chunks(self, parser, pagination, it_total, instance_id, limit):
        """Fetch added rows from this object instance from the Intelligence
        Server."""
        with tqdm(
            desc="Downloading",
            total=it_total + 1,
            disable=not self._progress_bar or not config.verbose,
            delay=3,
        ) as fetch_pbar:
            fetch_pbar.update()
            for _offset in range(self._initial_limit, pagination['total'], limit):
                response = self.__get_chunk(
                    instance_id=instance_id, offset=_offset, limit=limit
                )
                fetch_pbar.update()
                fetch_pbar.set_postfix(
                    rows=str(min(_offset + limit, pagination['total']))
                )
                parser.parse(response=response.json())

    def __initialize_report(self, limit: int) -> requests.Response:
        inst_pbar = tqdm(
            desc='Initializing an instance of a report. Please wait...',
            bar_format='{desc}',
            leave=False,
            ncols=285,
            disable=not self._progress_bar or not config.verbose,
        )

        # Switch off subtotals if I-Server version is higher than 11.2.1
        body = self._filter._request_body()
        if meets_minimal_version(self._connection.iserver_version, "11.2.0100"):
            self._subtotals["visible"] = False
            body["subtotals"] = {"visible": self._subtotals["visible"]}

        if self._current_page_by:
            body["currentPageBy"] = [{"id": el} for el in self._current_page_by]

        # Request a new instance, set instance id
        response = reports_api.report_instance(
            connection=self._connection,
            report_id=self._id,
            project_id=self.project_id,
            body=body,
            offset=0,
            limit=limit or self._initial_limit,
        )
        inst_pbar.close()
        return response

    def __get_chunk(
        self, instance_id: str, offset: int, limit: int
    ) -> requests.Response:
        return reports_api.report_instance_id(
            connection=self._connection,
            report_id=self._id,
            instance_id=instance_id,
            offset=offset,
            limit=limit,
        )

    def apply_filters(
        self,
        attributes: list | None = None,
        metrics: list | None = None,
        attr_elements: list | None = None,
        operator: str = 'In',
    ) -> None:
        """Apply filters on the reports's objects.

        Filter by attributes, metrics and attribute elements.

        Args:
            attributes (list or None, optional): ids of attributes to be
                included in the filter. If list is empty, no attributes will be
                selected and metric data will be aggregated.
            metrics (list or None, optional): ids of metrics to be included in
                the filter. If list is empty, no metrics will be selected.
            attr_elements (list or None, optional): attribute elements to be
                included in the filter.
            operator (str, optional): a str flag used to specify if the
                attribute elements selected inside the filter should be included
                or excluded. Allowed values are: 'In', 'NotIn'.
        """
        filtering_is_requested = bool(
            not all(element is None for element in [attributes, metrics, attr_elements])
        )

        if self._cross_tab:
            self._cross_tab_filter = {
                'attributes': attributes,
                'metrics': metrics,
                'attr_elements': attr_elements,
            }
        elif filtering_is_requested:
            self._filter._clear(
                attributes=attributes, metrics=metrics, attr_elements=attr_elements
            )
            self._filter.operator = operator
            self._select_attribute_filter_conditionally(attributes)
            self._select_metric_filter_conditionally(metrics)
            self._select_attr_el_filter_conditionally(attr_elements)
            # Clear instance, to generate new with new filters
            self._instance_id = None

    def _select_attribute_filter_conditionally(self, attributes_filtered) -> None:
        if attributes_filtered:
            self._filter._select(object_id=attributes_filtered)
        elif attributes_filtered is not None:
            self._filter.attr_selected = []

    def _select_metric_filter_conditionally(self, metrics_filtered) -> None:
        if metrics_filtered:
            self._filter._select(object_id=metrics_filtered)
        elif metrics_filtered is not None:
            self._filter.metr_selected = []

    def _select_attr_el_filter_conditionally(self, attr_el_filtered) -> None:
        if attr_el_filtered is not None:
            self._filter._select_attr_el(element_id=attr_el_filtered)

    def clear_filters(self) -> None:
        """Clear previously set filters, allowing all attributes, metrics, and
        attribute elements to be retrieved."""

        self._filter._clear()
        if self._cross_tab:
            self._filter._select(object_id=[el['id'] for el in self.attributes])
            self._filter._select(object_id=[el['id'] for el in self.metrics])
        # Clear instance, to generate new with new filters
        self._instance_id = None

    def _get_definition(self) -> None:
        """Get the definition of a report, including attributes and metrics.

        Implements GET /v2/reports/<report_id>.
        """
        response = reports_api.report_definition(
            connection=self._connection, report_id=self._id
        ).json()

        grid = response["definition"]["grid"]
        available_objects = response['definition']['availableObjects']

        if meets_minimal_version(self._connection.iserver_version, "11.2.0100"):
            self._subtotals = grid["subtotals"]
        self.name = response["name"]
        self._cross_tab = grid["crossTab"]

        # Check if report has consolidations
        if available_objects['consolidations']:
            exception_handler(
                msg="Reports with consolidations are not supported.",
                exception_type=ImportError,
            )

        full_attributes = []
        full_page_by_attributes = []
        full_attributes.extend(
            row for row in grid["rows"] if row["type"] in ["attribute", "customgroup"]
        )
        full_attributes.extend(
            column
            for column in grid["columns"]
            if column["type"] in ["attribute", "customgroup"]
        )
        full_page_by_attributes.extend(
            paging
            for paging in grid["pageBy"]
            if paging["type"] in ["attribute", "customgroup"]
        )
        self._attributes = [
            {'name': attr['name'], 'id': attr['id'], "type": attr["type"]}
            for attr in full_attributes
        ]
        self._page_by_attributes = [
            {'name': attr['name'], 'id': attr['id'], "type": attr["type"]}
            for attr in full_page_by_attributes
        ]

        # Retrieve metrics from the report grid (only selected metrics)
        metrics_position = grid.get("metricsPosition")
        if metrics_position is None:
            self._metrics = []
        else:
            full_metrics = grid[metrics_position["axis"]][metrics_position["index"]][
                "elements"
            ]
            self._metrics = [
                {'name': metr['name'], 'id': metr['id']} for metr in full_metrics
            ]

        self.__definition_retrieved = True

    def _get_pure_attributes(self) -> list:
        """Get attributes strictly of attribute type,
        e.g. for element operations."""
        attrs = []
        if self.attributes:
            attrs += self.attributes
        if self.page_by_attributes:
            attrs += self.page_by_attributes
        attrs = [a for a in attrs if a["type"] == "attribute"]
        return attrs

    def __get_attr_elements(self, limit: int = 50000) -> list:
        """Get elements of report attributes synchronously.

        Implements GET /reports/<report_id>/attributes/<attribute_id>/elements.
        """

        def fetch_for_attribute(attribute):
            @fallback_on_timeout()
            def fetch_for_attribute_given_limit(limit):
                response = reports_api.report_single_attribute_elements(
                    connection=self._connection,
                    report_id=self._id,
                    attribute_id=attribute['id'],
                    offset=0,
                    limit=limit,
                )
                # Get total number of rows
                total = get_total_count_of_objects(response)
                # Get attribute elements from the response.
                elements = response.json()

                assert isinstance(limit, int) and limit > 0
                assert isinstance(total, int) and total >= 0

                # If total number of elements is bigger than the chunk size
                # (limit), fetch them incrementally.
                for _offset in range(limit, total, limit):
                    response = reports_api.report_single_attribute_elements(
                        connection=self._connection,
                        report_id=self._id,
                        attribute_id=attribute['id'],
                        offset=_offset,
                        limit=limit,
                    )
                    elements.extend(response.json())

                # Return attribute data.
                return {
                    "attribute_name": attribute['name'],
                    "attribute_id": attribute['id'],
                    "elements": elements,
                }

            return fetch_for_attribute_given_limit(limit)[0]

        attrs = self._get_pure_attributes()
        attr_elements = []
        pbar = tqdm(
            attrs,
            desc="Loading attribute elements",
            leave=False,
            disable=not self._progress_bar or not config.verbose,
        )
        attr_elements = [fetch_for_attribute(attribute) for attribute in pbar]

        return attr_elements

    def __get_attr_elements_async(self, limit: int = 50000) -> list:
        """Get elements of report attributes asynchronously.

        Implements GET /reports/<report_id>/attributes/<attribute_id>/elements.
        """

        attrs = self._get_pure_attributes()
        attr_elements = []

        if attrs:
            threads = get_parallel_number(len(attrs))
            with FuturesSessionWithRenewal(
                connection=self._connection, max_workers=threads
            ) as session:
                # Fetch first chunk of attribute elements.
                futures = self.__fetch_attribute_elements_chunks(session, limit)
                pbar = tqdm(
                    futures,
                    desc="Loading attribute elements",
                    leave=False,
                    disable=not self._progress_bar or not config.verbose,
                )
                for i, future in enumerate(pbar):
                    attr = attrs[i]
                    response = future.result()
                    if not response.ok:
                        response_handler(
                            response, f"Error getting attribute {attr['name']} elements"
                        )
                    elements = response.json()
                    # Get total number of rows
                    total = get_total_count_of_objects(response)
                    for _offset in range(limit, total, limit):
                        response = reports_api.report_single_attribute_elements(
                            connection=self._connection,
                            report_id=self._id,
                            attribute_id=attr["id"],
                            offset=_offset,
                            limit=limit,
                        )
                        elements.extend(response.json())
                    # Append attribute data to the list of attributes.
                    attr_elements.append(
                        {
                            "attribute_name": attr['name'],
                            "attribute_id": attr['id'],
                            "elements": elements,
                        }
                    )
                pbar.close()

        return attr_elements

    def __fetch_attribute_elements_chunks(self, future_session, limit: int) -> list:
        # Fetch add'l rows from this object instance
        return [
            reports_api.report_single_attribute_elements_coroutine(
                future_session,
                report_id=self._id,
                attribute_id=attribute['id'],
                offset=0,
                limit=limit,
            )
            for attribute in self._get_pure_attributes()
        ]

    def list_properties(self):
        """List all properties of the object."""

        attributes = {
            key: self.__dict__[key] for key in self.__dict__ if not key.startswith('_')
        }
        attributes = {
            **attributes,
            "id": self.id,
            "instance_id": self.instance_id,
            "type": self.type,
            "subtype": self.subtype,
            "ext_type": self.ext_type,
            "date_created": self.date_created,
            "date_modified": self.date_modified,
            "version": self.version,
            "owner": self.owner,
            "view_media": self.view_media,
            "ancestors": self.ancestors,
            "certified_info": self.certified_info,
            "acg": self.acg,
            "acl": self.acl,
            "attributes": self.attributes,
            "metrics": self.metrics,
        }
        return {
            key: attributes[key]
            for key in sorted(attributes, key=key_fn_for_sort_object_properties)
        }

    def list_available_schedules(
        self, to_dictionary: bool = False
    ) -> list["Schedule"] | list[dict]:
        """Get a list of schedules available for the report.

        Args:
            to_dictionary (bool, optional): If True returns a list of
                dictionaries, otherwise returns a list of Schedules.
                False by default.

        Returns:
            List of Schedule objects or list of dictionaries.
        """
        schedules_list_response = (
            get_contents_schedule(
                connection=self.connection,
                project_id=self.connection.project_id,
                body={'id': self.id, 'type': 'report'},
            )
            .json()
            .get('schedules')
        )
        if to_dictionary:
            return schedules_list_response
        return [
            Schedule.from_dict(connection=self.connection, source=schedule_id)
            for schedule_id in schedules_list_response
        ]

    def get_selected_page_by_elements(self, valid_combination: list[int]) -> list[str]:
        """Get the selected page-by elements based on the valid combination.

        Args:
            valid_combination (list[int]): A list of valid element indexes. Can
                be obtained as one element from the `valid_page_by_elements`
                property.

        Returns:
            list[str]: A list of selected page-by element IDs.
        """
        selected_elements = []
        if len(valid_combination) != len(self.page_by_attributes):
            raise ValueError(
                "The length of valid_combination does not match the number of "
                "page-by attributes."
            )
        for index, attr in enumerate(self.page_by_attributes):
            if valid_combination[index] < len(self.page_by_elements[attr['id']]):
                selected_elements.append(
                    self.page_by_elements[attr['id']][valid_combination[index]]
                )
            else:
                raise IndexError(
                    f"Index {valid_combination[index]} is out of range for "
                    f"attribute {attr['name']} with "
                    f"{len(self.page_by_elements[attr['id']])} elements."
                )

        return selected_elements

    def get_valid_page_by_elements(self) -> dict:
        """Get a list of valid page-by elements for the report. It contains all
        elements of all attributes for page-by. Also includes possible
        combinations of selections for page-by.

        Returns:
            dict: A dictionary containing:
                - 'page_by_elements': All elements of all page-by attributes
                - 'valid_page_by_elements': Valid combinations of page-by
                    selections
        """

        if not self.instance_id:
            res = self.__initialize_report(self._initial_limit)
            _instance = res.json()
            self._instance_id = _instance['instanceId']

        # TODO: After solving this issue: CGPY-2482 to accept -1 as no limit
        # for this endpoint, change limit to -1 to get all elements at once.

        initial_response = reports_api.get_available_page_by_elements(
            self.connection,
            self.project_id,
            self.id,
            self._instance_id,
        )

        elements = camel_to_snake(initial_response.json())
        page_by_elements = elements.get('page_by', {})
        if self._page_by_elements is None:
            self._page_by_elements = {}
        self._page_by_elements.update(
            # convert to terse format for compatibility with Attributes API
            {
                attr['id']: [el['id'] for el in attr['elements']]
                for attr in elements.get('page_by', [])
            }
        )
        self._valid_page_by_elements = elements.get('valid_page_by_elements', {}).get(
            'items', []
        )

        return {
            'page_by_elements': page_by_elements,
            'valid_page_by_elements': self._valid_page_by_elements,
        }

    @property
    def attributes(self):
        if not self.__definition_retrieved:
            self._get_definition()
        return self._attributes

    @property
    def metrics(self):
        if not self.__definition_retrieved:
            self._get_definition()
        return self._metrics

    @property
    def page_by_attributes(self):
        if not self.__definition_retrieved:
            self._get_definition()
        return self._page_by_attributes

    @property
    def attr_elements(self):
        """Elements of attributes selected in the report. Custom Groups are
        excluded."""
        if not self.__definition_retrieved:
            self._get_definition()
        if not self._attr_elements and self._id:
            if self._parallel is True:
                # TODO: move the fallback inside the function to apply
                # per-attribute, like with non-async version.
                self._attr_elements = fallback_on_timeout()(
                    self.__get_attr_elements_async
                )(50000)[0]
            else:
                self._attr_elements = self.__get_attr_elements()
            self._filter._populate_attr_elements(self._attr_elements)
        return self._attr_elements

    @property
    def page_by_elements(self):
        """Elements of attributes selected for Page By in the report.
        The IDs are in the terse format used for page selection.
        """
        if any(
            attr['id'] not in self._page_by_elements for attr in self.page_by_attributes
        ):
            self.get_valid_page_by_elements()
        return self._page_by_elements

    @property
    def valid_page_by_elements(self) -> list:
        """Valid combinations of page by elements. It is a list of lists,
        where each inner list represents a valid combination of element
        selections for the page-by attributes. If there are N page-by
        attributes, each inner list will contain N element indexes,
        corresponding to the elements of each attribute."""
        if not self._valid_page_by_elements:
            self.get_valid_page_by_elements()
        return self._valid_page_by_elements

    @property
    def current_page_by(self):
        """Selected page of the report broken down by selected attributes."""
        return self._current_page_by

    @property
    def _filter(self):
        if not self.__definition_retrieved:
            self._get_definition()
        if self.__filter is None:
            self.__filter = Filter(
                attributes=self._attributes,
                metrics=self._metrics,
                attr_elements=self._attr_elements,
            )
        return self.__filter

    @property
    def selected_attributes(self):
        """Selected attributes for filtering."""
        return self._filter.attr_selected

    @property
    def selected_metrics(self):
        """Selected metrics for filtering."""
        return self._filter.metr_selected

    @property
    def selected_attr_elements(self):
        """Selected attribute elements for filtering."""
        return self._filter.attr_elem_selected

    @property
    def dataframe(self) -> pd.DataFrame:
        if self._dataframe is None:
            exception_handler(
                msg="Dataframe not loaded. Retrieve with Report.to_dataframe().",
                exception_type=Warning,
            )
        return self._dataframe

    @property
    def sql(self) -> str:
        """SQL View of the Report."""
        if self._sql is None:
            temp_instance_id = (
                self._instance_id
                if self._instance_id
                else reports_api.report_instance(
                    connection=self.connection,
                    report_id=self.id,
                    execution_stage='resolve_prompts',
                )
                .json()
                .get('instanceId')
            )
            self._sql = (
                reports_api.report_sql(
                    connection=self._connection,
                    report_id=self.id,
                    instance_id=temp_instance_id,
                )
                .json()
                .get('sqlStatement')
            )
        return self._sql

    @property
    def instance_id(self):
        return self._instance_id

    @instance_id.setter
    def instance_id(self, value):
        self._instance_id = value



# Test all datasource connections.

from mstrio import config
from mstrio.connection import get_connection, Connection
from mstrio.datasources import list_datasource_instances, DatasourceConnection

config.verbose = False

# Suppress default logging if verbose output is disabled
if not config.verbose:
    for hdlr in config.logger.handlers:
        hdlr.setLevel(50)


logger = config.logger


def test_all_datasources(conn: Connection, show_embedded: bool = False) -> None:
    datasources = list_datasource_instances(connection=conn)

    for ds in datasources:
        if isinstance(ds.datasource_connection, DatasourceConnection):
            try:
                test_results = ds.datasource_connection.test_connection()
                logger.info(
                    f"{'Success' if test_results else 'Failure'}  {ds.name}; DBC: "
                    f"'{ds.datasource_connection.name}'; DBL: "
                    f"'{ds.datasource_connection.datasource_login.name}' "
                    f"({ds.datasource_connection.datasource_login.username})"
                )
            except Exception as e:
                logger.error(
                    f"[Error]: {ds.name}; DBC: '{ds.datasource_connection.name}': "
                    f"{str(e) if config.verbose else str(e).splitlines()[0]}"
                )
        else:
            if show_embedded:
                logger.info(
                    f"[Warning]: No test {ds.name}; Embedded connection, test "
                    "not available"
                )


conn = get_connection(workstationData)

test_all_datasources(conn=conn, show_embedded=True)



"""This is the demo script to show how administrator can manage users and
user groups.

User management is showcased in the first part of the snippet and user group management in the second.
Additionally, the final section explains alternative method for user authentication and how to manage user's addresses.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""
import csv
from mstrio.connection import get_connection
from mstrio.access_and_security import Privilege, PrivilegeMode
from mstrio.users_and_groups import (
    create_users_from_csv, list_user_groups, list_users, User, UserGroup
)

# User management
# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to
PROJECT_ID = $project_id # Project id to connect to
NAME_BEGINS = $name_begins # Beginning of the full name to look for in listing users
ABBREVIATION_BEGINS = $abbreviation_begins # Beginning of the user abbreviation
USERNAME_0 = $username_0
FULLNAME_0 = $fullname_0
USER_ID_0 = $user_id_0
USER_ID_1 = $user_id_1
USER_ID_2 = $user_id_2
OWNER_1 = $owner_name

# Create connection based on workstation data
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# List users
list_users(
    connection=conn,
    name_begins=NAME_BEGINS,
    abbreviation_begins=ABBREVIATION_BEGINS,
    to_dictionary=False, # If True returns dict, by default (False) returns User objects.
    id=[USER_ID_1, USER_ID_2],  # Different filtering methods are available in the docstring.
)
# Get user by name
user = User(connection=conn, name=USERNAME_0)
# Get user by username
user = User(connection=conn, username=FULLNAME_0)
# Get user by id
user = User(connection=conn, id=USER_ID_0)

# List user properties
user_properties = user.list_properties()
print(user_properties)
# List user privileges
user.list_privileges(mode=PrivilegeMode.ALL, to_dataframe=False) # If True, return a DataFrame object containing privileges.
user.list_privileges(mode=PrivilegeMode.INHERITED, to_dataframe=True)

# Define variables which can be later used in a script
USERNAME_1 = $username_1
FULLNAME_1 = $full_name_1
USERNAME_2 = $username_2
FULLNAME_2 = $full_name_2
USERNAME_3 = $username_3
FULLNAME_3 = $full_name_3
USERNAME_4 = $username_4
FULLNAME_4 = $full_name_4
USERNAME_5 = $username_5
FULLNAME_5 = $full_name_5

# Note: To create user with no password you must set minimum password length as
# 0 in environment security settings
# create multiple users
users_array = [
    {
        'username': USERNAME_1, 'fullName': FULLNAME_1
    }, {
        'username': USERNAME_2, 'fullName': FULLNAME_2
    }, {
        'username': USERNAME_3, 'fullName': FULLNAME_3
    }, {
        'username': USERNAME_4, 'fullName': FULLNAME_4
    }, {
        'username': USERNAME_5, 'fullName': FULLNAME_5
    }
]
for u in users_array:
    User.create(connection=conn, username=u['username'], full_name=u['fullName'])

# Define a variable which can be later used in a script
CSV_FILE = $csv_file  # csv file to create user from

# Also, you can create users from a CSV file
newly_created_users = create_users_from_csv(connection=conn, csv_file=CSV_FILE)

# Note: To create user with no password you must set minimum password length as
# 0 in environment security settings
# Or you can do it manually
with open(CSV_FILE, "r") as f:
    users = csv.DictReader(f)

    for user in users:
        User.create(connection=conn, username=user['username'], full_name=user['full_name'])

# Define variables which can be later used in a script
USERNAME_6 = $username_6
FULLNAME_6 = $full_name_6
INITIALS = $initials # initials to look for in listing users

# Note: To create user with no password you must set minimum password length as
# 0 in environment security settings
# create a single user without password and get users which name begins with
# "John" and have additional filter for initials
User.create(connection=conn, username=USERNAME_6, full_name=FULLNAME_6)
my_users = list_users(connection=conn, name_begins=NAME_BEGINS, initials=INITIALS)

# Define variables which can be later used in a script
USERNAME_7 = $username_7
FULLNAME_7 = $full_name_7
PASSWORD_7 = $password_7
PRIVILEGE_NAME = $privilege_name
PRIVILEGE_ID = $privilege_id
FOLDER_ID = $folder_id
DESTINATION_FOLDER_PATH = $destination_folder_path
NEW_USERNAME = $new_username
NEW_LANGUAGE = $new_language
NEW_DEFAULT_TIMEZONE = $new_default_timezone
NEW_COMMENTS = $new_comments_for_user_object

# Create a single user with password already defined
user = User.create(connection=conn, username=USERNAME_7, full_name=FULLNAME_7, password=PASSWORD_7)
# Create profile folder for user
user.create_profile_folder(
    destination_folder=FOLDER_ID, # Folder class instance or ID as string, name as string or path as string specifying the folder where to create object can be provided.
    project=PROJECT_ID # It can be Project class instance or ID as string or name as string.
)

# Alter single user properties, for example change username, language and default timezone along with comment.
user.alter(
    username=NEW_USERNAME,
    language=NEW_LANGUAGE,
    default_timezone=NEW_DEFAULT_TIMEZONE,
    comments=NEW_COMMENTS
)
# List all possible privileges
all_privileges = Privilege.list_privileges(conn)
# Get Privilege object by name
priv = Privilege(conn, name=PRIVILEGE_NAME)
# Get Privilege object by id
priv = Privilege(conn, id=PRIVILEGE_ID)
# Grant privilege to user
user.grant_privilege(priv)
# Grant multiple privileges to user
user.grant_privilege([PRIVILEGE_ID, PRIVILEGE_NAME])
# Revoke privilege from user
user.revoke_privilege(PRIVILEGE_ID)
# Revoke multiple privileges from user
user.revoke_privilege([PRIVILEGE_ID, PRIVILEGE_NAME])

# Delete single user
user.delete(force=True, delete_profile=True)  # If delete profile is set to True, User's profile folder will be deleted as well.

# User Group Management
# Define a variable which can be later used in a script
USER_GROUP_NAME = $user_group_name  # name of a User Group to create and add users
USER_GROUP_ID = $user_group_id
USER_GROUP_1 = $user_group_1
USER_GROUP_ID_1 = $user_group_id_1
USER_GROUP_2 = $user_group_2
USER_GROUP_ID_2 = $user_group_id_2
DATE_CREATED = $date_created  # YYYY-MM-DD format
GROUP_NAME = $group_name
GROUP_DESCRIPTION = $group_description
USER_GROUP_LDAPDN = $user_group_ldapdn

# Get user group by name
user_group = UserGroup(connection=conn, name=USER_GROUP_NAME)
# also you can get user group by id
user_group = UserGroup(connection=conn, id=USER_GROUP_ID)
# Get all user groups (you can also add additional filters as for users)
user_groups_list = list_user_groups(
    connection=conn,
    name_begins=GROUP_NAME,
    description=GROUP_DESCRIPTION
)

# Get all members of group (you can also add additional filters as for users/user groups)
user_group.list_members(
    users=[user.id],
    date_created=DATE_CREATED,
)

# Get user group privileges
user_group.list_privileges(
    mode=PrivilegeMode.ALL, # Returns user group's inherited, granted or all privileges
)
# if instead of UserGroup class instance you would like to get object containing privileges, you can use to_dataframe flag
user_group.list_privileges(
    mode=PrivilegeMode.ALL,
    to_dataframe=False # # If True, return a DataFrame object containing privileges.
)
# Get user group properties
user_group_properties = user_group.list_properties()
print(user_group_properties)

# Create new user group
user_group = UserGroup.create(
    connection=conn,
    name=USER_GROUP_NAME,
    members=[USER_ID_0, USER_ID_1, USER_ID_2], # Specify members of the group
    ldapdn=USER_GROUP_LDAPDN, # Specify UserGroup ldapdn.
    description=GROUP_DESCRIPTION # Provide UserGroup description.
)
# Add user to user group
user_group.add_users(users=[user.id])
# Grant privilege to user group
user_group.grant_privilege(PRIVILEGE_ID)
# Grant multiple privileges to user group
user_group.grant_privilege([PRIVILEGE_ID, PRIVILEGE_NAME])
# Add user group to already existing groups
user_group.add_to_user_groups(groups=[USER_GROUP_ID_1, USER_GROUP_ID_2])
# also you can provide UserGroup Instances instead of IDs
user_group.add_to_user_groups(
    groups=[
        UserGroup(connection=conn, name=USER_GROUP_1),
        UserGroup(connection=conn, name=USER_GROUP_2),
    ]
)

# Revoke privilege from user group
user_group.revoke_privilege(PRIVILEGE_ID)
# Remove user group from passed groups
user_group.remove_from_user_groups(groups=[USER_GROUP_ID_1, USER_GROUP_ID_2])
# also you can provide UserGroup Instances instead of IDs
user_group.remove_from_user_groups(
    groups=[
        UserGroup(connection=conn, name=USER_GROUP_1),
        UserGroup(connection=conn, name=USER_GROUP_2),
    ]
)
# Delete a user group
user_group.delete(force=True) # If True, then no additional prompt will be shown before deleting.

# To create API Token for a specific user to log in as them
# or provide them alternative way to authenticate, you need a
# "Create and Edit Users and User Groups" privilege
# and you create the API Token by using the below method:
api_token = user.get_api_token()

# Available values for permissions are 'grant', 'deny', 'default' or None
grant_right = 'right'

# Define variables which can be later used in a script
OBJECT_ID = $object_id  # ID of an object to which permissions will be set
OBJECT_TYPE = $object_type  # type of an object to which permissions will be set;
# see ObjectTypes enum in types.py for available object type values

# set custom permissions of the user for given objects
user.set_custom_permissions(
    to_objects=[OBJECT_ID],
    object_type=OBJECT_TYPE,
    execute=grant_right,
    use=grant_right,
    control=grant_right,
    delete=grant_right,
    write=grant_right,
    read=grant_right,
    browse=grant_right
)

# Addresses
ADDRESS_NAME = $address_name
ADDRESS = $address
user_john = User(connection=conn, name=FULLNAME_6)
# The user's addresses list is lazily loaded upon first accessing the property
johns_addresses = user_john.addresses
# Strategy One allows having multiple addresses marked as default
# as long as they are assigned to different device.
# The 'default' argument is automatically set to True.
# Let's add new default address to John.
user_john.add_address(name=ADDRESS_NAME, address=ADDRESS)
# It can also be achieved by setting value of 'default' argument to True.
user_john.add_address(name=ADDRESS_NAME, address=ADDRESS, default=True)
# If the address should not be default, the 'default' should be set to False.
user_john.add_address(name=ADDRESS_NAME, address=ADDRESS, default=False)
# Let's construct a list of John's "default" addresses
johns_default_addresses = [addr for addr in johns_addresses if addr["is_default"]]
# If we just want a single address to treat as default and we're OK with
# it being just the first "default" one we can now just take the first element
# of the list
johns_default_address = johns_default_addresses[0]
# Alternatively if we don't care about any addresses besides the first
# "default one" we can use a generator expression and take its first output
# to avoid going through the entire list, increasing performance
johns_default_address = next((addr for addr in user_john.addresses if addr["is_default"]), None)
# By default, when only ADDRESS_NAME and ADDRESS are provided, an email type
# address is created. It is possible to create more address types by providing
# DEVICE_ID and DELIVERY_TYPE
ADDRESS_NAME = $address_name
ADDRESS = $address
DEVICE_ID = $device_id
DELIVERY_TYPE = $delivery_type
user_john.add_address(
    name=ADDRESS_NAME,
    address=ADDRESS,
    device_id=DEVICE_ID,
    delivery_type=DELIVERY_TYPE
)
# It is possible to make changes to an existing address
ADDRESS_ID = $address_id
NEW_ADDRESS_NAME = $new_address_name
NEW_ADDRESS = $new_address
NEW_DEVICE_ID = $new_device_id
NEW_DELIVERY_TYPE = $new_delivery_type
user_john.update_address(
    id=ADDRESS_ID,
    name=NEW_ADDRESS_NAME,
    address=NEW_ADDRESS,
    device_id=NEW_DEVICE_ID,
    delivery_type=NEW_DELIVERY_TYPE
)
# It is possible to copy address from another user
FULLNAME_8 = $full_name_8
user_with_address = User(connection=conn, name=FULLNAME_8)
user_john.add_address(contact_address=user_with_address.addresses[0])



import logging
import uuid
from enum import auto
from typing import TYPE_CHECKING

from mstrio import config
from mstrio.api import custom_groups
from mstrio.modeling.expression import Expression, ExpressionFormat
from mstrio.modeling.metric.metric_format import MetricFormat
from mstrio.modeling.schema.helpers import ObjectSubType
from mstrio.object_management import Folder, SearchPattern, full_search
from mstrio.object_management.object import Object
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import AutoName, get_enum_val
from mstrio.utils.helper import (
    Dictable,
    construct_expression_body,
    delete_none_values,
    filter_params_for_func,
    find_object_with_name,
)
from mstrio.utils.resolvers import (
    get_drill_map_id_from_params_set,
    get_folder_id_from_params_set,
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import custom_groups as cg_processors
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import (
    class_version_handler,
    meets_minimal_version,
    method_version_handler,
)
from mstrio.utils.vldb_mixin import ModelVldbMixin

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


class CustomGroupOptions(Dictable):
    """Custom Group display options.

    Attributes:
        hierarchy_display: Whether to enable hierarchical display. Default is
            True.
        subtotals_display: Whether to enable subtotals display. Default is
            False.
        element_header_above_child: Whether to display element header above
            child elements. Default is True.
    """

    def __init__(
        self,
        hierarchy_display: bool | None = None,
        subtotals_display: bool | None = None,
        element_header_above_child: bool | None = None,
    ):
        self.hierarchy_display = hierarchy_display
        self.subtotals_display = subtotals_display
        self.element_header_above_child = element_header_above_child


class ElementDisplayOption(AutoName):
    """Enumeration of Custom Group Element display options.

    Attributes:
        ELEMENT: Display only the element name.
        INDIVIDUAL: Display only individual items within the element.
        INDIVIDUAL_EXPAND: Display individual items within the element. Expand
            individual items if possible.
        ELEMENT_INDIVIDUAL_EXPAND: Display both the element name and individual
            items. Expand individual items if possible.
    """

    ELEMENT = auto()
    INDIVIDUAL = auto()
    INDIVIDUAL_EXPAND = auto()
    ELEMENT_INDIVIDUAL_EXPAND = auto()


class CustomGroupElement(Dictable):
    """Custom Group Element representation.

    Attributes:
        name (str): Name of the custom group element.
        display (ElementDisplayOption): Display option for the element.
        element_format (MetricFormat): Format settings for the element.
        items_format (MetricFormat): Format settings for individual items in
            the report.
        qualification (Expression): Qualification expression for the element.
        id (str, optional): ID of the custom group element. If not provided,
            it will be generated automatically and replaced by server-generated
            ID when Custom Group is created on the server.
    """

    _FROM_DICT_MAP = {
        "display": ElementDisplayOption,
        "element_format": MetricFormat.from_dict,
        "items_format": MetricFormat.from_dict,
        "qualification": Expression.from_dict,
    }

    def __init__(
        self,
        name: str,
        display: ElementDisplayOption,
        qualification: Expression,
        element_format: MetricFormat | None = None,
        items_format: MetricFormat | None = None,
        id: str | None = None,
    ):
        self.id = id or str(uuid.uuid4())
        self.name = name
        self.display = display
        self.element_format = element_format
        self.items_format = items_format
        self.qualification = qualification


@method_version_handler("11.3.0200")
def list_custom_groups(
    connection: "Connection",
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project: "Project | str | None" = None,
    project_id: str | None = None,
    project_name: str | None = None,
    search_pattern: SearchPattern | int = SearchPattern.CONTAINS,
    show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
    folder: "Folder | tuple[str] | list[str] | str | None" = None,
    folder_id: str | None = None,
    folder_name: str | None = None,
    folder_path: tuple[str] | list[str] | str | None = None,
    **filters,
) -> list["CustomGroup"] | list[dict]:
    """Get a list of CustomGroup objects or dicts. Optionally filter the
    objects by specifying `filters` parameters.

    Args:
        connection (object): Strategy One connection object returned by
            `connection.Connection()`
        name (str, optional): value the search pattern is set to, which
            will be applied to the names of CustomGroup objects being searched
        to_dictionary (bool, optional): If `True` returns dictionaries, by
            default (`False`) returns `CustomGroup` objects.
        limit (int, optional): limit the number of elements returned. If `None`
            (default), all objects are returned.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        search_pattern (SearchPattern enum or int, optional): pattern to
            search for, such as Begin With or Exactly. Possible values are
            available in ENUM `mstrio.object_management.SearchPattern`.
            Default value is CONTAINS (4).
        show_expression_as (ExpressionFormat or str, optional): specify how
            expressions should be presented
            Available values:
            - None
            (expression is returned in "text" format)
            - `ExpressionFormat.TREE` or `tree`
            (expression is returned in `text` and `tree` formats)
            - `ExpressionFormat.TOKENS` or `tokens`
            (expression is returned in `text` and `tokens` formats)
        folder (Folder | tuple | list | str, optional): Folder object or ID or
            name or path specifying the folder. May be used instead of
            `folder_id`, `folder_name` or `folder_path`.
        folder_id (str, optional): ID of a folder.
        folder_name (str, optional): Name of a folder.
        folder_path (str, optional): Path of the folder.
            The path has to be provided in the following format:
                if it's inside of a project, start with a Project Name:
                    /MicroStrategy Tutorial/Public Objects/Metrics
                if it's a root folder, start with `CASTOR_SERVER_CONFIGURATION`:
                    /CASTOR_SERVER_CONFIGURATION/Users
        **filters: Available filter parameters: ['id', 'name',
            'type', 'subtype', 'date_created', 'date_modified', 'version',
            'acg', 'owner', 'ext_type']
    Returns:
        list of filter objects or list of filter dictionaries.
    """
    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    validate_owner_key_in_filters(filters)

    objects = full_search(
        connection,
        object_types=ObjectSubTypes.CUSTOM_GROUP,
        project=proj_id,
        pattern=search_pattern,
        name=name,
        root=folder,
        root_id=folder_id,
        root_name=folder_name,
        root_path=folder_path,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects

    # `show_expression_as` does not apply to search. It has to be stored in the
    # `CustomGroup` object for when the element expressions are retrieved later
    # in the object's lifetime.
    return [
        CustomGroup.from_dict(
            source={
                "show_expression_as": (
                    show_expression_as
                    if isinstance(show_expression_as, ExpressionFormat)
                    else ExpressionFormat(show_expression_as)
                ),
                **obj,
            },
            connection=connection,
        )
        for obj in objects
    ]


@class_version_handler("11.3.0200")
class CustomGroup(Entity, CopyMixin, MoveMixin, DeleteMixin, ModelVldbMixin):
    """Python representation of a Custom Group object.

    Attributes:
        name (str): Name of the custom group.
        id (str): ID of the custom group.
        description (str): Description of the custom group.
        sub_type (ObjectSubType): Subtype of the object in metadata, here:
            CUSTOM_GROUP.
        version (str): Version of the object in metadata.
        ancestors (list[dict]): List of ancestor folders.
        type (ObjectTypes): Type of the object in metadata, here: FILTER.
        date_created (datetime): Date and time when the object was created.
        date_modified (datetime): Date and time when the object was last
            modified.
        owner (dict): Information about the owner of the object.
        acg (Rights): Access rights flags for the object.
        acl (list[ACE]): Access control list for the object.
        is_embedded (bool): If True, indicates that the target object of this
            reference is embedded within this object. Alternatively if this
            object is itself embedded, then it means that the target object is
            embedded in the same container as this object. If this field is
            omitted (as is usual) then the target is not embedded.
        path (str): Path to the object in the project/folder structure.
        primary_locale (str): The primary locale of the object, in the
            IETF BCP 47 language tag format, such as "en-US". Read only.
        hidden (bool): Whether the object is hidden in the folder structure.
        elements (list[CustomGroupElement]): List of custom group elements.
            Note: IDs of the elements are generated by the server independently
            from the client-side generated IDs and differ between fetches.
        drill_map (dict, optional): Drill map used by the custom group.
        options (CustomGroupOptions, optional): Display options for the custom
            group.
    """

    _OBJECT_TYPE = ObjectTypes.FILTER
    _OBJECT_SUBTYPES = [ObjectSubTypes.CUSTOM_GROUP]

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        "owner": User.from_dict,
        "subtype": ObjectSubTypes,
        "sub_type": ObjectSubType,
        "elements": [CustomGroupElement.from_dict],
        "options": CustomGroupOptions.from_dict,
    }
    _API_GETTERS = {
        (
            "type",
            "subtype",
            "ext_type",
            "date_created",
            "date_modified",
            "version",
            "owner",
            "ancestors",
            "acg",
            "acl",
            "hidden",
            "comments",
        ): objects_processors.get_info,
        (
            "id",
            "name",
            "description",
            "sub_type",
            "date_created",
            "date_modified",
            "path",
            "version_id",
            "is_embedded",
            "primary_locale",
            "destination_folder_id",
            "elements",
            "drill_map",
            "options",
        ): cg_processors.get_custom_group,
    }
    _API_PATCH = {
        (
            "name",  # required by Custom Group REST even if not changing
            "primary_locale",
            "elements",
            "drill_map",
            "options",
        ): (cg_processors.update_custom_group, "partial_put"),
        (
            "description",
            "folder_id",
            "hidden",
            "comments",
            "owner",
        ): (objects_processors.update, "partial_put"),
    }

    _MODEL_VLDB_API = {
        'GET_ADVANCED': custom_groups.get_custom_group,
        'GET_APPLICABLE': custom_groups.get_applicable_vldb_settings,
    }

    def __init__(
        self,
        connection: "Connection",
        id: str | None = None,
        name: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        show_filter_tokens: bool = False,
    ):
        """Initialize Custom Group object by its identifier.

        Args:
            connection: Strategy One connection object returned
                by `connection.Connection()`
            id (str, optional): Metadata ID of the Custom Group. May be omitted
                if `name` parameter is provided.
            name (str, optional): Name of the Custom Group.
            show_expression_as (enum or str, optional): specify how expressions
                should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS` or `tokens`
            show_filter_tokens (bool, optional): specify whether `qualification`
                is returned in `tokens` format, along with `text` and `tree`
                format
        """
        connection._validate_project_selected()
        if id is None:
            if name is None:
                raise ValueError(
                    "Please specify either 'name' or 'id' parameter in the constructor."
                )

            found_cg = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_custom_groups,
            )
            id = found_cg["id"]

        super().__init__(
            connection=connection,
            object_id=id,
            show_expression_as=show_expression_as,
            show_filter_tokens=show_filter_tokens,
        )

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)
        self.primary_locale = kwargs.get("primary_locale", default_value)
        self.is_embedded = kwargs.get("is_embedded", default_value)
        self.destination_folder_id = kwargs.get("destination_folder_id", default_value)
        self._sub_type = (
            ObjectSubType(kwargs.get("sub_type"))
            if kwargs.get("sub_type")
            else default_value
        )
        self._path = kwargs.get("path", default_value)
        self.elements = (
            [CustomGroupElement.from_dict(el) for el in kwargs.get("elements", [])]
            if kwargs.get("elements")
            else default_value
        )
        self.drill_map = kwargs.get("drill_map", default_value)
        self.options = (
            CustomGroupOptions.from_dict(kwargs.get("options"))
            if kwargs.get("options")
            else default_value
        )
        show_expression_as = kwargs.get("show_expression_as", "tree")
        self._show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )

        # PUT endpoint requires 'information.name' in the body even if name is
        # not being changed. Wrapper must know the objects body, hence setting
        # method here.
        self._MODEL_VLDB_API = self._MODEL_VLDB_API.copy()
        self._MODEL_VLDB_API["PUT_ADVANCED"] = self._update_vldb_properties

    @staticmethod
    def _construct_element_body(element: CustomGroupElement) -> dict:
        body = element.to_dict()
        body["qualification"] = construct_expression_body(element.qualification)
        return body

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        elements: list[CustomGroupElement],
        destination_folder: "Folder | tuple[str] | list[str] | str | None" = None,
        destination_folder_path: tuple[str] | list[str] | str | None = None,
        description: str | None = None,
        primary_locale: str | None = None,
        options: CustomGroupOptions | None = None,
        drill_map: Object | str | None = None,
        drill_map_id: str | None = None,
        drill_map_name: str | None = None,
        show_expression_as: ExpressionFormat | str = ExpressionFormat.TREE,
        journal_comment: str | None = None,
    ) -> "CustomGroup":
        """Create a new Custom Group object on the server.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            name (str): Name of the custom group.
            elements (list[CustomGroupElement]): List of elements to include in
                the custom group.
            destination_folder (Folder | tuple | list | str, optional): Folder
                object or ID or name or path specifying the folder where to
                create object.
            destination_folder_path (str, optional): Path of the folder.
                The path has to be provided in the following format:
                    /MicroStrategy Tutorial/Public Objects/Metrics
            description (str, optional): Description of the custom group.
            primary_locale (str, optional): the primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US"
            options (CustomGroupOptions, optional): Display options
                for the custom group.
            drill_map (Object | str, optional): Object or ID or name specifying
                the drill map. May be used instead of `drill_map_id` or
                `drill_map_name`.
            drill_map_id (str, optional): ID of a drill map.
            drill_map_name (str, optional): Name of a drill map.
            show_expression_as (ExpressionFormat | str, optional): How
                expressions should be presented. Defaults to
                ExpressionFormat.TREE.
            journal_comment (optional, str): Comment that will be added to the
                object's change journal entry

        Returns:
            CustomGroup | dict: Newly created Custom Group object.
        """

        dest_id = get_folder_id_from_params_set(
            connection,
            connection.project_id,
            folder=destination_folder,
            folder_path=destination_folder_path,
        )
        drill_map_id = get_drill_map_id_from_params_set(
            connection,
            drill_map=drill_map,
            drill_map_id=drill_map_id,
            drill_map_name=drill_map_name,
        )
        body = {
            "information": {
                "name": name,
                "description": description,
                "destinationFolderId": dest_id,
                "primaryLocale": primary_locale,
            },
            "drillMap": {"objectId": drill_map_id} if drill_map_id else None,
            "options": options.to_dict() if options else None,
            "elements": [cls._construct_element_body(el) for el in elements],
        }

        if journal_comment:
            body.update({"changeJournal": {"userComments": journal_comment}})
        body = delete_none_values(body, recursion=True)

        response = cg_processors.create_custom_group(
            connection,
            body=body,
            project_id=connection.project_id,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
        )

        if config.verbose:
            logger.info(
                f"Successfully created custom group named: '{name}' with ID: '"
                f"{response['id']}'"
            )
        return cls.from_dict(
            source={
                **response,
                "show_expression_as": show_expression_as,
            },
            connection=connection,
        )

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        primary_locale: str | None = None,
        elements: list[CustomGroupElement] | None = None,
        options: CustomGroupOptions | None = None,
        drill_map: Object | str | None = None,
        drill_map_id: str | None = None,
        drill_map_name: str | None = None,
        remove_drill_map: bool = False,
        hidden: bool | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
        journal_comment: str | None = None,
    ):
        """Alter properties of the Custom Group object.
        Args:
            name (str, optional): New name of the custom group.
            description (str, optional): New description of the custom group.
            primary_locale (str, optional): New primary locale of the object,
                in the IETF BCP 47 language tag format, such as "en-US".
            elements (list[CustomGroupElement], optional): New list of elements
                to include in the custom group. This will completely replace
                the existing list of elements.
            options (CustomGroupOptions, optional): New display options
                for the custom group. May be partially specified to update only
                certain options.
            drill_map (Object | str, optional): Object or ID or name specifying
                the drill map. May be used instead of `drill_map_id` or
                `drill_map_name`.
            drill_map_id (str, optional): ID of a drill map.
            drill_map_name (str, optional): Name of a drill map.
            remove_drill_map (bool, optional): If `True`, removes the drill map
                from the custom group. Defaults to `False`.
            hidden (bool, optional): Whether the object is hidden in the folder
                structure.
            comments (str, optional): Long description for the object.
            owner (str | User, optional): New owner of the object. Can be
                specified either as User object or user ID.
            journal_comment (str, optional): Comment that will be added to the
                object's change journal entry.
        """
        if name is None:
            name = self.name  # REST requires name in body, even if not changing
        drill_map_id = get_drill_map_id_from_params_set(
            self.connection,
            drill_map=drill_map,
            drill_map_id=drill_map_id,
            drill_map_name=drill_map_name,
        )
        if drill_map_id:
            drill_map = {"objectId": drill_map_id}
        if remove_drill_map:
            drill_map = {}
            if not meets_minimal_version(self.connection.iserver_version, "11.6.0100"):
                logger.warning(
                    "Removing Drill Map from Custom Group is not supported "
                    "in I-Server versions prior to 11.6.0100. The operation "
                    "will be attempted, but may fail."
                )
        if elements:
            elements = [self._construct_element_body(el) for el in elements]
        if isinstance(owner, User):
            owner = owner.id

        if remove_drill_map:
            # When fetching afterwards, "drill_map" field will be unset rather
            # than set to None and the field in local object will not update
            # If failed to remove, will fetch again
            self.drill_map = None

        properties = filter_params_for_func(
            self.alter,
            locals(),
            exclude=["self", "drill_map_id", "drill_map_name", "remove_drill_map"],
        )
        properties = delete_none_values(properties, recursion=True)
        self._alter_properties(**properties)

    @staticmethod
    def _match_element(
        el: CustomGroupElement, id: str | None, name: str | None
    ) -> bool:
        """Match Custom Group Element by ID or name query args. To be used by
        element-related methods.

        Args:
            el (CustomGroupElement): Custom Group Element to match against.
            id (str, optional): ID of the custom group element.
            name (str, optional): Name of the custom group element.

        Returns:
            bool: True if the element matches the query, False otherwise.
        """
        if id is not None:
            return el.id == id
        if name is not None:
            return el.name == name
        return False

    def get_element(
        self,
        id: str | None = None,  # must be one of the two
        name: str | None = None,
        to_dictionary: bool = False,
    ) -> CustomGroupElement | dict:
        """Get a specific Custom Group Element by its ID or name.

        Note:
            Provide either `id` or `name` parameter, but not both.

        Args:
            id (str, optional): ID of the custom group element. May be omitted
                if `name` parameter is provided.
            name (str, optional): Name of the custom group element.
            to_dictionary (bool, optional): If `True` returns a dictionary
                instead of a `CustomGroupElement` object. Default is `False`.

        Returns:
            CustomGroupElement | dict | None: The specified custom group element
                or None if not found.
        """
        if not bool(id) ^ bool(name):
            raise ValueError(
                "Please specify either `id` or `name`, but not both nor neither"
            )
        for el in self.elements:
            if self._match_element(el, id, name):
                return el.to_dict() if to_dictionary else el
        return None

    def create_element(
        self,
        name: str,
        display: ElementDisplayOption,
        qualification: Expression,
        element_format: MetricFormat | None = None,
        items_format: MetricFormat | None = None,
        journal_comment: str | None = None,
    ):
        """Create a new Custom Group Element and add it to the Custom Group.

        Args:
            name (str): Name of the custom group element.
            display (ElementDisplayOption): Display option for the element.
            element_format (MetricFormat, optional): Format settings for
                the element.
            items_format (MetricFormat, optional): Format settings for
                individual items in the report.
            qualification (Expression): Qualification expression
                for the element.
            journal_comment (str, optional): Comment that will be added to the
                object's change journal entry.
        """
        new_element = CustomGroupElement(
            name=name,
            display=display,
            element_format=element_format,
            items_format=items_format,
            qualification=qualification,
        )
        new_elements = self.elements.copy()
        new_elements.append(new_element)
        self.alter(elements=new_elements, journal_comment=journal_comment)

    def alter_element(
        self,
        id: str,
        name: str | None = None,
        display: ElementDisplayOption | None = None,
        element_format: MetricFormat | None = None,
        items_format: MetricFormat | None = None,
        reset_format: bool = False,
        qualification: Expression | None = None,
        journal_comment: str | None = None,
    ):
        """Alter properties of a specific Custom Group Element.

        Args:
            id (str): ID of the custom group element to alter.
            name (str, optional): New name of the custom group element.
            display (ElementDisplayOption, optional): New display option
                for the element.
            element_format (MetricFormat, optional): Format settings for
                the element.
            items_format (MetricFormat, optional): Format settings for
                individual items in the report.
            reset_format (bool, optional): If `True`, erase format settings
                previously set on the server for both `element_format` and
                `items_format`. Defaults to `False`.
            qualification (Expression, optional): New qualification expression
                for the element.
            journal_comment (str, optional): Comment that will be added to the
                object's change journal entry.
        """
        altered_element = self.get_element(id=id)
        if altered_element is None:
            if config.verbose:
                logger.warning(f"No element found with ID: {id}")
            return
        if name is not None:
            altered_element.name = name
        if display is not None:
            altered_element.display = display
        if element_format is not None or reset_format:
            altered_element.element_format = element_format
        if items_format is not None or reset_format:
            altered_element.items_format = items_format
        if qualification is not None:
            altered_element.qualification = qualification
        new_elements = [el for el in self.elements if not el.id == id]
        new_elements.append(altered_element)
        self.alter(elements=new_elements, journal_comment=journal_comment)

    def delete_element(
        self,
        id: str,
        journal_comment: str | None = None,
    ):
        """Delete a specific Custom Group Element from the Custom Group.

        Args:
            id (str): ID of the custom group element to delete.
            journal_comment (str, optional): Comment that will be added to the
                object's change journal entry.
        """
        new_elements = [el for el in self.elements if not el.id == id]
        if len(new_elements) == len(self.elements):
            if config.verbose:
                logger.warning(f"No element found with ID: {id}")
        else:
            self.alter(elements=new_elements, journal_comment=journal_comment)

    # VLDB methods. Wrapped with version checks as relevant APIs were added
    # later than the ones for Custom Group object.

    @method_version_handler("11.5.0900")
    def list_vldb_settings(
        self,
        set_names=None,
        names=None,
        groups=None,
        to_dictionary=False,
        to_dataframe=False,
    ):
        return super().list_vldb_settings(
            set_names, names, groups, to_dictionary, to_dataframe
        )

    @method_version_handler("11.5.0900")
    def alter_vldb_settings(self, names_to_values):
        return super().alter_vldb_settings(names_to_values)

    @method_version_handler("11.5.0900")
    def reset_vldb_settings(self, set_names=None, names=None, groups=None, force=False):
        return super().reset_vldb_settings(set_names, names, groups, force)

    def _update_vldb_properties(self, connection: "Connection", id: str, body: dict):
        """Update VLDB properties of the Custom Group object.
        Wraps custom_groups.update_custom_group API call to account for
            required 'information.name' field in the body.
        """
        if "information" not in body:
            body["information"] = {}
        if "name" not in body["information"]:
            body["information"]["name"] = self.name
        custom_groups.update_custom_group(connection, id, body)

    # TODO: remove to standalone function when implementing Drill Map class
    @staticmethod
    def _list_drill_maps(connection: "Connection"):
        return full_search(
            connection,
            object_types=ObjectTypes.DRILL_MAP,
            project=connection.project_id,
        )



"""
    This is a template for creation of Simple Python Function which can be
    called in a Metric expression.

    It will work out of the box as is, when all variables are filled in.

    Method `get_results` needs to exist and return the values served as the
    metric evaluation results.

    Metric expression example: ScriptFunctionSimple<_InputNames="x, y",
                               _ScriptName=”minus_two_numbers”>(Revenue, Cost)
"""

"""
    `x` is a scalar input
    `y` is a scalar input
"""

def minus_two_numbers(x, y):
    return x - y

# must-have method which returns the values served as the metric evaluation results. 
def get_results():
    return minus_two_numbers($x, $y)

print(get_results())



"""This is the demo script to show how to search for items using mstrio-py with
limited data available.

This script will not work without replacing parameters with real values.
"""

from mstrio.connection import get_connection
from mstrio.object_management.search_operations import find_objects_with_id

# Define variables which can be later used in a script
OBJECT_ID = $object_id  # ID of an object to search for
SEARCH_PROJECT_ID1 = $search_project_id1
SEARCH_PROJECT_ID2 = $search_project_id2

# Connect to the Strategy One environment
conn = get_connection(workstationData)

# There might be many cases where you have access to an ID of an object but not
# to its type, project or projects (if any) where it is.
#
# mstrio-py has a method to find the object(s) in such cases and you can use it
# as follows:
elements = find_objects_with_id(conn, OBJECT_ID)

# [DISCLAIMER]
# Have in mind this is a very resource intensive method and should be used for
# debugging purposes and not in a production code unless necessary.

# The returned `elements` is a list of objects with a provided `OBJECT_ID` in
# a particular shape:
#
# [
#   {"project_id": <str>, "object_data": <Entity-based-object>},
#   ...
# ]
#
# If an object is a configuration-level object (for example a `User`),
# the returned list would look like this (notice `project_id` being `None` for
# configuration-level objects):
#
# [
#     {
#         "project_id": None,
#         "object_data": User(conn, id=OBJECT_ID),
#     }
# ]
#
# If an object is a project-level object (for example a `Folder`), all instances
# of it will be found and the list may look like this:
#
# [
#   {"project_id": "B3FEE61A11E696...",
#    "object_data": Folder(conn, id=OBJECT_ID)},
#   {"project_id": "4BAE16A340B995...",
#    "object_data": Folder(conn, id=OBJECT_ID)},
#   {"project_id": "0DDDDEC8C94B32...",
#    "object_data": Folder(conn, id=OBJECT_ID)},
#   {"project_id": "B7CA92F04B9FAE...",
#    "object_data": Folder(conn, id=OBJECT_ID)},
#   {"project_id": "4C09350211E697...",
#    "object_data": Folder(conn, id=OBJECT_ID)},
#   {"project_id": "CE52831411E696...",
#    "object_data": Folder(conn, id=OBJECT_ID)},
# ]

# If you know that a particular ID can be found only in at least one of specific
# places, you can provide additional parameters to the method to narrow the
# search.
#
# If you know that the object is a configuration-level object:
elements = find_objects_with_id(
    conn, OBJECT_ID, config_level_only=True
)

# If you know that the object is only within at most a set of
# particular Projects:
elements = find_objects_with_id(
    conn, OBJECT_ID, projects=[SEARCH_PROJECT_ID1, SEARCH_PROJECT_ID2]
)

# Additionally, if you wish to receive only a raw dictionary of data for each
# found object instead of an initialized mstrio-py's module, you can provide
# a flag to do so:
elements = find_objects_with_id(conn, OBJECT_ID, to_dictionary=True)
